#!/usr/bin/python2
# -*- coding: utf-8 -*-

import sqlite3 as lite
import os
import math

try:
    from process.error_handling import KillByUser
except ImportError:
    from trifusion.process.error_handling import KillByUser


"""my @steps = ( # Common
       ['updateMinimumEvalueExponent'],
       ['bestQueryTaxonScore'],
       ['qtscore_ix'],

       # Ortholog
       ['bestHit'],
       ['best_hit_ix'],
       ['ortholog', ["drop table BestHit"]],
       ['orthologTaxon'],
       ['orthologAvg'],
       ['orthologAvgIndex'],
       ['orthologsNormalization', ["drop table OrthologAvgScore", "drop table OrthologTaxon", "drop table OrthologTemp"]],

       # InParalog
       ['bestInterTaxonScore', ["drop table BestQueryTaxonScore"]],
       ['bis_uids_ix'],
       ['uniqueSimSeqsQueryId'],
       ['ust_qids_ix'],
       ['betterHit', ["drop table BestInterTaxonScore", "drop table UniqSimSeqsQueryId"]],
       ['better_hit_ix'],
       ['inParalog', ["drop table BetterHit"]],
       ['inParalogTaxonAvg'],
       ['orthologUniqueId'],
       ['orthologUniqueIdIndex'],
       ['inplgOrthTaxonAvg', ["drop table OrthologUniqueId"]],
       ['inParalogAvg',["drop table InParalogTaxonAvg", "drop table InplgOrthTaxonAvg"]],
       ['inParalogAvgIndex'],
       ['inParalogsNormalization', ["drop table InParalogAvgScore", "drop table InParalogTemp"]],

       # CoOrtholog
       ['inParalog2Way'],
       ['in2a_ix'],
       ['in2b_ix'],
       ['ortholog2Way'],
       ['ortholog2WayIndex'],
       ['inplgOrthoInplg'],
       ['inParalogOrtholog'],
       ['coOrthologCandidate', ["drop table Ortholog2Way", "drop table InParalog2Way", "drop table InplgOrthoInplg", "drop table InParalogOrtholog"]],
       ['coOrthologNotOrtholog', ["drop table CoOrthologCandidate"]],
       ['coOrthologNotOrthologIndex'],
       ['coOrtholog', ["drop table CoOrthNotOrtholog"]],
       ['coOrthologTaxon'],
       ['coOrthologAvg'],
       ['coOrthologAvgIndex'],
       ['coOrthologsNormalization', ["drop table CoOrthologAvgScore", "drop table CoOrthologTaxon", "drop table CoOrthologTemp"]],
       ['cleanall', ["truncate table InParalog", "truncate table Ortholog", "truncate table CoOrtholog"]],
      );
"""
################################################################################
############################### Auxiliar    #################################
################################################################################

def log(value):
    return math.log10(value)

def orthologTaxonSub (cur, co):

    #assuming in perl a var is true if not ""
    coCaps = "" if co == "" else "Co"
    t1 = coCaps + 'OrthologTaxon'
    t2 = coCaps + 'OrthologTemp'

    cur.execute("create table %s as\
        select case\
        when taxon_id_a < taxon_id_b\
        then taxon_id_a\
        else taxon_id_b\
        end as smaller_tax_id,\
        case\
        when taxon_id_a < taxon_id_b\
        then taxon_id_b\
        else taxon_id_a\
        end as bigger_tax_id,\
        unnormalized_score\
        from %s " % (t1, t2))

#from (?) ", (coCaps + 'OrthologTaxon', coCaps + 'OrthologTemp'))

def normalizeOrthologsSub (cur, co, table):

    #assuming in perl a var is true if not ""
    coCaps = "" if co == "" else "Co"
    co = "o" if co == "" else "coO"
    t1 = coCaps + 'OrthologAvgScore'
    t2 = coCaps + 'OrthologTaxon'

    cur.execute("create table %s as\
        select smaller_tax_id, bigger_tax_id, avg(unnormalized_score) avg_score\
        from %s\
        group by smaller_tax_id, bigger_tax_id" % (t1, t2))

################################################################
    t1 = co+'orthoAvg_ix'
    t2 = coCaps+'OrthologAvgScore'

    cur.execute("create unique index %s on %s(smaller_tax_id,bigger_tax_id,avg_score)" % (t1, t2))

################################################################

    t1 = coCaps + 'OrthologTemp'
    t2 = coCaps + 'OrthologAvgScore'

    cur.execute("insert into %s (sequence_id_a, sequence_id_b, taxon_id_a, taxon_id_b, unnormalized_score, normalized_score)\
        select ot.sequence_id_a, ot.sequence_id_b, ot.taxon_id_a, ot.taxon_id_b, ot.unnormalized_score, ot.unnormalized_score/a.avg_score\
        from %s ot, %s a\
        where min(ot.taxon_id_a, ot.taxon_id_b) = a.smaller_tax_id\
        and max(ot.taxon_id_a, ot.taxon_id_b) = a.bigger_tax_id" % (table, t1, t2))


################################################################################
############################### Common tables #################################
################################################################################
def commonTempTables (cur):


    cur.execute("select min(evalue_exp)\
        from SimilarSequences\
        where evalue_mant != 0")

    tup = cur.fetchone()
    minEvalueExp = tup[0] - 1

    cur.execute("update SimilarSequences\
        set evalue_exp = ?\
        where evalue_exp = 0 and evalue_mant = 0", (minEvalueExp,))

##########################################################################

    cur.execute("create table BestQueryTaxonScore as\
        select im.query_id as query_id, im.subject_taxon_id as subject_taxon_id, low_exp.evalue_exp as evalue_exp, min(im.evalue_mant) as evalue_mant\
        from InterTaxonMatch im,\
        (select query_id, subject_taxon_id, min(evalue_exp) as evalue_exp\
        from InterTaxonMatch\
        group by query_id, subject_taxon_id) low_exp\
        where im.query_id = low_exp.query_id\
        and im.subject_taxon_id = low_exp.subject_taxon_id\
        and im.evalue_exp = low_exp.evalue_exp\
        group by im.query_id, im.subject_taxon_id, low_exp.evalue_exp")

################################################################################

    cur.execute("create unique index qtscore_ix on BestQueryTaxonScore(query_id, subject_taxon_id, evalue_exp, evalue_mant)")

################################################################################
############################### Orthologs #####################################
################################################################################
def orthologs(cur):

    cur.execute("create table BestHit as\
        select s.query_id, s.subject_id,\
        s.query_taxon_id, s.subject_taxon_id,\
        s.evalue_exp, s.evalue_mant\
        from SimilarSequences s, BestQueryTaxonScore cutoff\
        where s.query_id = cutoff.query_id\
        and s.subject_taxon_id = cutoff.subject_taxon_id\
        and s.query_taxon_id != s.subject_taxon_id\
        and s.evalue_exp <= -5\
        and s.percent_match >= 50\
        and (s.evalue_mant < 0.01\
        or s.evalue_exp = cutoff.evalue_exp\
        and s.evalue_mant = cutoff.evalue_mant)")

    cur.execute("create unique index best_hit_ix on BestHit(query_id,subject_id)")

######################################################################

    cur.execute("create table OrthologTemp as\
        select bh1.query_id as sequence_id_a, bh1.subject_id as sequence_id_b,\
        bh1.query_taxon_id as taxon_id_a, bh1.subject_taxon_id as taxon_id_b,\
        case\
        when bh1.evalue_mant < 0.01 or bh2.evalue_mant < 0.01\
        then (bh1.evalue_exp + bh2.evalue_exp) / -2\
        else\
        (log(bh1.evalue_mant * bh2.evalue_mant)\
        + bh1.evalue_exp + bh2.evalue_exp) / -2\
        end as unnormalized_score\
        from BestHit bh1, BestHit bh2\
        where bh1.query_id < bh1.subject_id\
        and bh1.query_id = bh2.subject_id\
        and bh1.subject_id = bh2.query_id")

######################################################################

    orthologTaxonSub(cur, '')

######################################################################

    normalizeOrthologsSub(cur, '', "Ortholog")


################################################################################
############################### InParalogs ####################################
################################################################################
def inparalogs (cur):

    cur.execute("create table BestInterTaxonScore as\
        select im.query_id as query_id, low_exp.evalue_exp as evalue_exp, min(im.evalue_mant) as evalue_mant\
        from BestQueryTaxonScore im,\
        (select query_id, min(evalue_exp) as evalue_exp\
        from BestQueryTaxonScore\
        group by query_id) low_exp\
        where im.query_id = low_exp.query_id\
        and im.evalue_exp = low_exp.evalue_exp\
        group by im.query_id, low_exp.evalue_exp")

###########################################################################

    cur.execute("create unique index bis_uids_ix on BestInterTaxonScore(query_id)")

###########################################################################

    cur.execute("create table UniqSimSeqsQueryId as\
        select distinct s.query_id from SimilarSequences s")

###########################################################################

    cur.execute("create unique index ust_qids_ix on UniqSimSeqsQueryId (query_id)")

###########################################################################

    cur.execute("CREATE TABLE BetterHit as\
        select s.query_id, s.subject_id,\
        s.query_taxon_id as taxon_id,\
        s.evalue_exp, s.evalue_mant\
        from SimilarSequences s, BestInterTaxonScore bis\
        where s.query_id != s.subject_id \
        and s.query_taxon_id = s.subject_taxon_id\
        and s.query_id = bis.query_id\
        and s.evalue_exp <= -5\
        and s.percent_match >= 50\
        and (s.evalue_mant < 0.001\
        or s.evalue_exp < bis.evalue_exp\
        or (s.evalue_exp = bis.evalue_exp and s.evalue_mant <= bis.evalue_mant))\
        union\
        select s.query_id, s.subject_id, s.query_taxon_id as taxon_id, s.evalue_exp, s.evalue_mant\
        from SimilarSequences s\
        where s.query_taxon_id = s.subject_taxon_id \
        and s.evalue_exp <= -5\
        and s.percent_match >= 50\
        and s.query_id in \
        (SELECT distinct ust.query_id\
        from UniqSimSeqsQueryId ust\
        LEFT OUTER JOIN BestInterTaxonScore bis ON bis.query_id = ust.query_id\
        WHERE bis.query_id IS NULL)")

###########################################################################

    cur.execute("create index better_hit_ix on BetterHit (query_id,subject_id)")

###########################################################################

    cur.execute("create table InParalogTemp as\
        select bh1.query_id as sequence_id_a, bh1.subject_id as sequence_id_b,\
        bh1.taxon_id,\
        case\
        when bh1.evalue_mant < 0.01 or bh2.evalue_mant < 0.01\
        then (bh1.evalue_exp + bh2.evalue_exp) / -2\
        else\
        (log(bh1.evalue_mant * bh2.evalue_mant)\
        + bh1.evalue_exp + bh2.evalue_exp) / -2\
        end as unnormalized_score\
        from BetterHit bh1, BetterHit bh2\
        where bh1.query_id < bh1.subject_id\
        and bh1.query_id = bh2.subject_id\
        and bh1.subject_id = bh2.query_id")

################################################################

    cur.execute("create table InParalogTaxonAvg as\
        select avg(i.unnormalized_score) average, i.taxon_id as taxon_id\
        from InParalogTemp i\
        group by i.taxon_id")

################################################################

    cur.execute("create table OrthologUniqueId as\
        select distinct(sequence_id) from (\
        select sequence_id_a as sequence_id from Ortholog\
        union\
        select sequence_id_b as sequence_id from Ortholog) i")

################################################################

    cur.execute("create unique index ortho_uniq_id_ix on OrthologUniqueId (sequence_id)")

################################################################

    cur.execute(" create table InplgOrthTaxonAvg as\
        select avg(i.unnormalized_score) average, i.taxon_id as taxon_id\
        from InParalogTemp i\
        where i.sequence_id_a in\
        (select sequence_id from OrthologUniqueId)\
        or i.sequence_id_b in\
        (select sequence_id from OrthologUniqueId)\
        group by i.taxon_id")

################################################################

    cur.execute("create table InParalogAvgScore as\
        select case\
        when orth_i.average is NULL\
        then all_i.average\
        else orth_i.average\
        end as avg_score,\
        all_i.taxon_id\
        from InParalogTaxonAvg all_i LEFT OUTER JOIN InplgOrthTaxonAvg orth_i\
        ON all_i.taxon_id = orth_i.taxon_id")

################################################################

    cur.execute("create unique index inparalog_avg_ix on InParalogAvgScore(taxon_id,avg_score)")

################################################################

    cur.execute(" insert into InParalog(sequence_id_a, sequence_id_b, taxon_id, unnormalized_score, normalized_score)\
        select it.sequence_id_a, it.sequence_id_b, it.taxon_id, it.unnormalized_score, it.unnormalized_score/a.avg_score\
        from InParalogTemp it, InParalogAvgScore a\
        where it.taxon_id = a.taxon_id")

################################################################################
############################### CoOrthologs ###################################
################################################################################
def coorthologs (cur):

    cur.execute("create table InParalog2Way as\
        select sequence_id_a, sequence_id_b from InParalog\
        union\
        select sequence_id_b as sequence_id_a, sequence_id_a as sequence_id_b from InParalog")

######################################################################

    cur.execute("create unique index in2a_ix on InParalog2Way(sequence_id_a, sequence_id_b)")

######################################################################

    cur.execute("create unique index in2b_ix on InParalog2Way(sequence_id_b, sequence_id_a)")

######################################################################

    cur.execute("create table Ortholog2Way as\
        select sequence_id_a, sequence_id_b from Ortholog\
        union\
        select sequence_id_b as sequence_id_a, sequence_id_a as sequence_id_b from Ortholog")

######################################################################

    cur.execute("create unique index ortholog2way_ix on Ortholog2Way(sequence_id_a, sequence_id_b)")

######################################################################

    cur.execute("create table InplgOrthoInplg as\
        select ip1.sequence_id_a, ip2.sequence_id_b\
        from Ortholog2Way o, InParalog2Way ip2, InParalog2Way ip1\
        where ip1.sequence_id_b = o.sequence_id_a\
        and o.sequence_id_b = ip2.sequence_id_a")

##################################################################

    cur.execute("create table InParalogOrtholog as\
        select ip.sequence_id_a, o.sequence_id_b\
        from InParalog2Way ip, Ortholog2Way o\
        where ip.sequence_id_b = o.sequence_id_a")
 
##################################################################

    cur.execute("create table CoOrthologCandidate as\
        select distinct\
        min(sequence_id_a, sequence_id_b) as sequence_id_a,\
        max(sequence_id_a, sequence_id_b) as sequence_id_b\
        from (select sequence_id_a, sequence_id_b from InplgOrthoInplg\
        union\
        select sequence_id_a, sequence_id_b from InParalogOrtholog) t")
 
######################################################################

    cur.execute("create table CoOrthNotOrtholog as\
        SELECT cc.sequence_id_a, cc.sequence_id_b\
        FROM CoOrthologCandidate cc\
        LEFT OUTER JOIN Ortholog o\
        ON cc.sequence_id_a = o.sequence_id_a\
        AND cc.sequence_id_b = o.sequence_id_b\
        WHERE o.sequence_id_a IS NULL")

#####################################################################

    cur.execute("create index cno_ix on CoOrthNotOrtholog(sequence_id_a,sequence_id_b)")

######################################################################

    cur.execute("create table CoOrthologTemp as\
        select candidate.sequence_id_a, candidate.sequence_id_b,\
        ab.query_taxon_id as taxon_id_a, ab.subject_taxon_id as taxon_id_b,\
        case\
        when ab.evalue_mant < 0.00001 or ba.evalue_mant < 0.00001\
        then (ab.evalue_exp + ba.evalue_exp) / -2\
        else\
        (log(ab.evalue_mant * ba.evalue_mant)\
        + ab.evalue_exp + ba.evalue_exp) / -2\
        end as unnormalized_score\
        from SimilarSequences ab, SimilarSequences ba, CoOrthNotOrtholog candidate\
        where ab.query_id = candidate.sequence_id_a\
        and ab.subject_id = candidate.sequence_id_b\
        and ab.evalue_exp <= -5\
        and ab.percent_match >= 50\
        and ba.query_id = candidate.sequence_id_b\
        and ba.subject_id = candidate.sequence_id_a\
        and ba.evalue_exp <= -5\
        and ba.percent_match >= 50")
    
######################################################################

    orthologTaxonSub(cur, 'co')

######################################################################

    normalizeOrthologsSub(cur, "Co", "CoOrtholog")


def execute(db_dir, nm=None):
    con = lite.connect(os.path.join(db_dir, "orthoDB.db"))

    with con:

        if nm:
            if nm.stop:
                raise KillByUser("")
            nm.total = 4
            nm.counter = 0
            nm.msg = None

        con.create_function("log", 1, log)

        cur = con.cursor()

        for func in [commonTempTables, orthologs, inparalogs, coorthologs]:

            if nm:
                if nm.stop:
                    raise KillByUser("")
                nm.counter += 1

            func(cur)

if __name__ == '__main__':
    execute(".")


__author__ = "Fernando Alves and Diogo N. Silva"



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
#  Copyright 2012 Unknown <diogo@arch>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.


class ArgumentError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class OutputFormatError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class InputError(Exception):
    def __init__(self, value="Invalid input"):
        self.value = value

    def __str__(self):
        return repr(self.value)


class EmptyAlignment(Exception):
    def __init__(self, value="Alignment is empty"):
        self.value = value

    def __str__(self):
        return repr(self.value)


class MultipleSequenceTypes(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class EmptyData(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class InvalidSequenceType(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class DuplicateTaxa(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class KillByUser(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return  repr(self.value)


class SingleAlignment(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


__author__ = "Diogo N. Silva"



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
#  Copyright 2012 Unknown <diogo@arch>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import matplotlib
# Using the agg backend prevent a segmentation fault when quiting the
# application
matplotlib.use("agg")

import matplotlib.pyplot as plt
import matplotlib.cm as cm
from scipy.stats import spearmanr
import numpy as np
from itertools import chain
from collections import Counter
import seaborn as sns
import os

# Set of 10 easily distinguishable colors that will be used when generating plots
# to ensure consistency in color usage. In case more than 10 colors are required
# (ugh...) they will be randomly generated henceforth.

clr_list = [[0, .53, .66],  # light blue
            [1, .16, .16],     # light red
            [.21, .78, .22],  # light green
            [.83, .55, .37],  # light brown
            [.55, .37, .83],  # purple
            [1, .4, 0],     # light orange
            [.77, .77, .77],  # light grey
            [.78, .22, .22],  # dark red
            [.1, .1, .1],     # dark grey
            [0, .66, 0]]      # green

two_clr_list = [[0, .53, .66],
                [.62, .80, .85]]


class SetProps(object):
    """
    This decorator is use to automatically set several plot properties based
    on the arguments provided to the plotting functions. In this way,
    there is no need to write repetitive code in each function. The supported
    arguments are:

    .: ax_names: list, first element x-axis label; second element y-axis label
    .: title: string, title of the plot
    """

    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):

        plt.clf()
        plt.close()

        res = self.func(*args, **kwargs)

        # Set axis names
        if "ax_names" in kwargs:
            if kwargs["ax_names"][0]:
                plt.xlabel(kwargs["ax_names"][0])
            if kwargs["ax_names"][1]:
                plt.ylabel(kwargs["ax_names"][1])

        # Set title
        if "title" in kwargs:
            plt.title(kwargs["title"])

        return res


def _add_labels(ax_names):
    """
    Wrapper to the add labels routine
    :param ax_names: List of two items. [x-axis name, y-axis name]
    """

    if ax_names[0]:
        plt.xlabel(ax_names[0])

    if ax_names[1]:
        plt.ylabel(ax_names[1])


@SetProps
def scatter_plot(data, correlation=False, ax_names=None, table_header=None,
                 title=None):
    """
    Builds a scatter plot from a 2D data list. Also calculates the
    correlation coefficient is requested.
    :param data: 2D list, containing the x and y data points
    :param correlation: Boolean. If True, the spearman's rank correlation
    coefficient is also calculated
    :param ax_names: list, with two string elements for the x and y axis
    :param table_header: list, with table header. Each element represents a
    column
    """

    plt.rcParams["figure.figsize"] = (8, 6)

    # Use ggpot style
    plt.style.use("ggplot")

    fig, ax = plt.subplots()

    # Generate scatter plot
    ax.scatter(data[0], data[1], alpha=0.5, edgecolors=None)

    # Plot best fit line
    ax.plot(data[0], np.poly1d(np.polyfit(data[0], data[1], 1))(data[0]))

    # Set plot limits
    max1 = max(data[0])
    max2 = max(data[1])
    ax.set_xlim(0, max1 + (max1 * .1))
    ax.set_ylim(0, max2 + (max2 * .1))

    # Calculate spearman's rank correlation coefficient and add it to the plot
    if correlation:
        rho, pval = spearmanr(data[0], data[1])

        ax.text((ax.get_xlim()[1] - ax.get_xlim()[0]) * .05,
                ax.get_ylim()[1] - (ax.get_ylim()[1] * .05),
                r"$\rho$ = {}; p-value = {}".format(round(rho, 4),
                                                    round(pval, 4)))

    # Generate table data
    if table_header:
        table = [table_header]
    else:
        table = []

    for i, j in zip(*data):
        table.append([i, j])

    return fig, None, table


@SetProps
def bar_plot(data, labels=None, title=None, ax_names=None,
             lgd_list=None, reverse_x=False, table_header=None):
    """
    Builds simple bar plot from a data_list
    :param data: list with data to be plotted.
    :param labels: list with x axis labels
    :param title: string, plot title
    :param ax_names: list. Names of the axis [yaxis_name, xaxis_name]
    :param reverse_x: Boolean, determines whether the x-axis is reversed or not
    """

    if len(labels) > 10:
        plt.rcParams["figure.figsize"] = (len(labels) / 3, 6)
    else:
        plt.rcParams["figure.figsize"] = (8, 6)

    # Use ggpot style
    plt.style.use("ggplot")

    fig, ax = plt.subplots()

    # Create plot
    for i, d in enumerate(data):
        # Get color from 10 color list. If more than 10 colors are required,
        # randomly generate new ones
        if len(data) > 2:
            try:
                clr = clr_list[i]
            except IndexError:
                clr = np.random.rand(3, 1)
        else:
            clr = two_clr_list[i]

        # Create plot for first entry
        if d == data[0]:
            ax.bar(np.arange(len(d)), d, align="center", color=clr,
                   label=lgd_list[i] if lgd_list else None)
        # Add plots on top
        else:
            ax.bar(np.arange(len(d)), d, align="center", color=clr,
                   bottom=data[i - 1],
                   label=lgd_list[i] if lgd_list else None)

    # Set legend
    if lgd_list:
        lgd = plt.legend(bbox_to_anchor=(1, .5), loc="center left",
                         fancybox=True, shadow=True, framealpha=0.8)
    else:
        lgd = None

    # Set labels at the center of bars
    if labels:
        ax.set_xticks(np.arange(len(labels)))
        # Determine rotation and alignmnet from labels
        if max([len(x) for x in labels]) >= 3:
            ax.set_xticklabels(labels, ha="right", rotation=45)
        else:
            ax.set_xticklabels(labels, ha="center")

        # Setting the x range to avoid large spaces between the axis and bars
        plt.xlim([min(np.arange(len(labels))) - .5,
                  max(np.arange(len(labels))) + .5])

    # Invert x-axis so that higher taxa prevalence is shown in the left
    if reverse_x:
        plt.gca().invert_xaxis()

    # Generating table structure
    if table_header:
        table = [table_header]
    else:
        table = []

    for l, x in zip(labels, data[0]):
        table.append([l, x])

    return fig, lgd, table


@SetProps
def multi_bar_plot(data_list, labels=None, lgd_list=None, title=None,
                   ax_names=None):
    """
    General purpose multiple bar plot with custom layout. Returns a pyplot
    object.
    :param data_list: list with data to be plotted in list/array type in each
    entry. Data with two groups would be like [[1.23, .53], [1.55, .12]]
    :param labels: list, containing the labels
    :param lgd_list: list, The legend string for each data set in data_list
    """

    plt.rcParams["figure.figsize"] = (8, 6)

    # Use ggpot style
    plt.style.use("ggplot")

    fig, ax = plt.subplots()

    # Set margin between bar groups
    margin = 0.1

    # Determine bar group width according to the number of data lists
    w = (1. - 2. * margin) / len(data_list)
    # Determine max plt height to calculate text label space
    max_height = max((x for y in data_list for x in y))

    # Create bar plots
    for i, d in enumerate(data_list):
        # Get color from 10 color list. If more than 10 colors are required,
        # randomly generate new ones
        try:
            clr = clr_list[i]
        except IndexError:
            clr = np.random.rand(3, 1)

            # Determine position of xdata
        xdata = np.arange(len(d)) + margin + (i * w)
        bplt = ax.bar(xdata, d, w, color=clr, label=lgd_list[i])

    # Add labels
    if labels:
        ax.set_xticks(np.arange(len(labels)) + 0.5)
        ax.set_xticklabels(labels, rotation=45, ha="center", fontweight="bold")

    # Set label colors
    ax.tick_params(axis="x", colors=[.4, .4, .4])
    ax.tick_params(axis="y", colors=[.3, .3, .3])

    # Add legend
    lgd = plt.legend(bbox_to_anchor=(1, .5), loc="center left",
                     fancybox=True, shadow=True, framealpha=0.8)

    # Add text to the top of bar with the corresponding value
    for b in bplt:

        # If the bar width is large enough to accommodate the text with
        # horizontal orientation
        if w >= 0.8:
            xpos = b.get_x() + b.get_width() / 2
            ypos = b.get_height() * .90
            ax.text(xpos, ypos, b.get_height(), horizontalalignment="center",
                    weight='bold', color="white", size=14)
        else:
            if b.get_height() / max_height >= 0.2:
                xpos = b.get_x() + b.get_width() / 2 * 1.3
                ypos = b.get_height() * .95
                ax.text(xpos, ypos, b.get_height(),
                        horizontalalignment="center", color="white",
                        verticalalignment="top", weight='bold',
                        size=14, rotation=90)
            else:
                xpos = b.get_x() + b.get_width() / 2
                ypos = b.get_height() * 1.1
                ax.text(xpos, ypos, b.get_height(),
                        horizontalalignment="center", color="black",
                        verticalalignment="bottom", weight='bold',
                        size=14, rotation=90)

    # Automatically adjust figure size to accommodate all labels
    # plt.tight_layout()

    return fig, lgd


@SetProps
def interpolation_plot(data, title=None, ax_names=None):
    """
    Creates a black and white interpolation plot from data, which must consist
    of a 0/1 matrix for absence/presence of taxa in genes
    :param data: numpy array of variable shape.
    """

    plt.rcParams["figure.figsize"] = (8, 6)

    # Use ggpot style
    plt.style.use("ggplot")

    fig, ax = plt.subplots()

    # Setting the aspect ratio proportional to the data
    ar = float(len(data[0])) / float(len(data)) * .2

    ax.imshow(data, interpolation="none", cmap="Greys", aspect=ar)

    ax.grid(False)

    return fig, None, None


@SetProps
def stacked_bar_plot(data, labels, legend=None, table_header=None, title=None,
                     ax_names=None, normalize=False, normalize_factor=None):
    """
    Creates a tight stacked bar plot
    :param data: list, data for 2 groups should be like [[1,2], [2,3]]
    :param labels: list, should match the number of items in data.
    """

    plt.rcParams["figure.figsize"] = (8, 6)

    if normalize:

        data_original = np.copy(data)
        data = np.array([[y / normalize_factor for y in x] for x in data])

    if len(labels) > 10:
        plt.rcParams["figure.figsize"] = (len(labels) / 3, 6)
    else:
        plt.rcParams["figure.figsize"] = (8, 6)

    # Use ggpot style
    plt.style.use("ggplot")

    fig, ax = plt.subplots()

    # Get bottom positions for stacked bar
    bottoms = np.cumsum(np.vstack((np.zeros(data.shape[1]), data)),
                        axis=0)[:-1]

    w = .8

    # Get x positions
    xpos = [x + (w / 2) for x in range(len(labels))]

    for c, d in enumerate(data):

        c1 = c if c < 14 else c - 14.5

        clr = cm.Set1(c1 / 14., 1)

        if c == 0:
            bplot = ax.bar(xpos, d, w, color=clr, label=legend[c])
        else:
            bplot = ax.bar(xpos, d, w, color=clr, label=legend[c],
                           bottom=bottoms[c])

    # Set x labels
    plt.xticks([x + (w / 2) for x in xpos], labels, ha="right", rotation=45)

    # Set legend
    if legend:
        cols = len(legend) if len(legend) < 6 else 6
        lgd = plt.legend(bbox_to_anchor=(0.5, 1.02), loc=8, fancybox=True,
                         shadow=True, framealpha=.8, ncol=cols)

    if ax_names:
        _add_labels(ax_names)

    # Generate table structure
    if table_header:
        table = [table_header]
    else:
        table = []
    if normalize:
        for i, lbl in enumerate(labels):
            table.append([lbl] + list(chain.from_iterable((int(x[i]), y[i])
                                      for x, y in zip(*[data_original, data]))))
    else:
        for i, lbl in enumerate(labels):
            table.append([lbl] + [int(x[i]) for x in data])

    return fig, lgd, table


@SetProps
def box_plot(data, labels=None, title=None, ax_names=None):
    """
    Creates a boxplot from a data series
    :param data: list, data to be plotted
    :param labels: list, x-axis labels
    :param title: string, plot title
    :param ax_names: list, first element for x-axis, second for y-axis
    """

    plt.rcParams["figure.figsize"] = (8, 6)

    if len(labels) > 10:
        plt.rcParams["figure.figsize"] = (len(labels) / 3, 6)
    else:
        plt.rcParams["figure.figsize"] = (8, 6)

    # Use ggpot style
    plt.style.use("ggplot")

    fig, ax = plt.subplots()

    bplot = ax.boxplot(data, patch_artist=True)

    # Change outline color, fill color and line width of the boxes
    for box in bplot["boxes"]:
        box.set(color="#7570b3", linewidth=2)
        box.set(facecolor="#1b9e77")

    # Change the color and line width of the whiskers
    for whisker in bplot["whiskers"]:
        whisker.set(color="#7570b3", linewidth=2)

    # Change color and line width of the caps
    for cap in bplot["caps"]:
        cap.set(color="#7570b3", linewidth=2)

    # Change color and line width of the medias
    for media in bplot["medians"]:
        media.set(color="#b2df8a", linewidth=2)

    # Change the style of the fliers and their fill
    for flier in bplot["fliers"]:
        flier.set(marker="o", color="#e7298a", alpha=0.5)

    if labels:
        ax.set_xticklabels(labels, rotation=45, ha="right")

    # Generate table structure
    table = [["", "1st Quartile", "Median", "3rd Quartile"]]

    for x, d in zip(labels, data):
        table.append([x, np.percentile(d, 25), np.median(d),
                      np.percentile(d, 75)])

    return fig, None, table


def histogram_smooth(data, ax_names=None, table_header=None, title=None,
                     legend=None):
    """
    Creates a smooth line plot with colored areas with the same distribution
    as an histogram
    :param data: list, with histogram data. To support multiple plots,
    provide each plot data as a list. If you want to produce two smooth
    histograms, provide a data list with two lists.
    :param title: string, title for the plot
    :param ax_names: list, names for the x-axis and y-axis, respectively
    :param table_header: list, header for the table
    """

    from scipy.interpolate import UnivariateSpline

    plt.rcParams["figure.figsize"] = (8, 6)

    plt.style.use("ggplot")

    fig, ax = plt.subplots(nrows=3, ncols=1, sharex=True, sharey=True)

    # Start building table
    if table_header:
        table = [table_header]
    else:
        table = []

    for (p, d), axn, l in zip(enumerate(data), ax, legend):
        # Convert data into histogram
        y, x = np.histogram(d, 50)
        x = x[:-1] + (x[1] - x[0]) / 2
        # Fit a one-dimensional smoothing spline to git a given set of points
        # from the histogram
        f = UnivariateSpline(x, y, s=200)
        # Set title for subplot
        axn.set_title(l, color=clr_list[p], fontweight="bold")
        # Create subplot
        axn.plot(x, f(x), color=clr_list[p])
        # Color area below line lot
        axn.fill_between(x, f(x), alpha=0.5, color=clr_list[p])

    # If axis names are provided, add them to figure
    if ax_names:
        fig.text(0.5, 0.02, ax_names[0], ha="center", fontweight="bold",
                 color="grey", fontsize=14)
        fig.text(0.04, 0.5, ax_names[1], va="center", rotation="vertical",
                 fontweight='bold', color="grey", fontsize=14)

    # Build histogram with fixed bin range for table construction
    hist_data = []
    for d in data:
        y, x = np.histogram(d, 100, (0., 1.))
        hist_data.append(y)

    # Populate table
    for i, d in zip(np.arange(0, 1, 0.01), zip(*hist_data)):
        table.append([i] + list(d))

    return fig, None, table


@SetProps
def histogram_plot(data, title=None, ax_names=None, table_header=None,
                   real_bin_num=False):
    """
    Creates an histogram from data
    :param data: list
    :param title: string
    :param ax_names: list, first element for x-axis, second for y-axis
    :param table_header: list, containing the header of the table as the
    first element
    :param real_bin_num: boolean, if true, the histogram bins will be real
    numbers
    """

    # Use ggpot style
    plt.style.use("ggplot")

    plt.rcParams["figure.figsize"] = (8, 6)

    fig, ax = plt.subplots()

    c_data = Counter(data)
    max_data = max(c_data)
    min_data = min(c_data)
    if len(c_data) > 50:
        bins = int(len(c_data) / 10)
    else:
        bins = len(c_data)

    vals, b, _ = plt.hist(data, bins, histtype="stepfilled",
                                color=clr_list[0])

    plt.axvline(np.mean(data), linewidth=2, color="r", alpha=.8,
                linestyle="--")

    # Add cutom artist for legend
    mean_artist = plt.Line2D((0, 1), (0, 1), color="r", linestyle="--")

    lgd = ax.legend([mean_artist], ["Mean"], loc=2, frameon=True, fancybox=True,
                    shadow=True, framealpha=.8, fontsize="large")
    lgd.get_frame().set_facecolor("white")

    # Generate table structure
    if table_header:
        table = [table_header]
    else:
        table = []

    if real_bin_num:
        # If real_bin_num was set, then the bins in the table should be real
        # numbers
        vals, b = np.histogram(data, bins, (min_data, max_data + 1))

        c = 0
        for b, v in zip(b, vals):
            table.append(["{} - {}".format(int(c), int(b)), v])
            c = b + 1

        table.append(["{} - {}".format(int(c), max_data), vals[-1]])

    else:
        for p, val in zip(b, vals):
            table.append([p, val])

    return fig, lgd, table


@SetProps
def triangular_heat(data, labels, color_label=None, title=None):
    """
    Creates a triangular heatmap plot based on an array of triangular shape,
    or with a masked triangle
    :return:
    """

    plt.style.use("ggplot")

    plt.rcParams["figure.figsize"] = (8, 6)

    fig, ax = plt.subplots()

    cmap = cm.get_cmap("jet", 100)
    cmap.set_bad("w")

    heat = ax.imshow(data, interpolation="nearest", cmap=cmap)

    # Set ticks for label positioning
    ax.set_xticks(np.arange(data.shape[1]), minor=False)
    ax.set_yticks(np.arange(data.shape[0]), minor=False)
    # Set minor ticks that will make the grid
    ax.set_xticks(np.arange(data.shape[1]) + .5, minor=True)
    ax.set_yticks(np.arange(data.shape[0]) - .5, minor=True)
    # Set x axis labels on top and y axis labels on right
    ax.xaxis.tick_top()
    # Set axis labels
    # Remove first entry of xlabel
    xlabel = ["" if x == 0 else labels[x] for x in range(len(labels))]
    ax.set_xticklabels(xlabel, rotation=45, ha="left")
    # Remove last entry of ylabel
    ylabel = ["" if x == len(labels) - 1 else labels[x] for x in
              range(len(labels))]
    ax.set_yticklabels(ylabel)
    # Remove major ticks
    plt.tick_params(axis="both", which="major", top="off", right="off",
                    left="off", labelsize=5)
    # Remove minor ticks
    plt.tick_params(axis="y", which="minor", right="off")

    plt.grid(True, which="minor")
    plt.grid(False, which="major")

    plt.gca().invert_xaxis()

    cbar = plt.colorbar(heat)
    if not color_label:
        cbar.set_label("Similarity proportion")
    else:
        cbar.set_label(color_label)

    # Generate table
    table = [[""] + [x for x in labels[::-1]]]
    for p, sp in enumerate(labels):
        table.append([sp] + list(data[p])[::-1])

    return fig, None, table


@SetProps
def outlier_densisty_dist(data, outliers, outliers_labels=None, ax_names=None,
                          title=None):
    """
    Creates a density distribution for data and highlights outliers
    :param data: 1D array containing data points
    :param outliers: 1D array containing the outliers
    :param outliers_labels: 1D array containing the name of the outliers.
    Must match outliers in size
    :param ax_names: list, with names for x and y axis, respectively
    """

    plt.rcParams["figure.figsize"] = (8, 6)

    fig, ax = plt.subplots()

    # Create density function
    sns.distplot(data, rug=True, hist=False, color="black")

    # Plot outliers
    ax.plot(outliers, np.zeros_like(outliers), "ro", clip_on=False,
            label="Outliers")

    # Create legend
    lgd = ax.legend(frameon=True, loc=2, fancybox=True, shadow=True,
                    framealpha=.8, prop={"weight": "bold"})

    if outliers_labels:
        table = [[os.path.basename(x)] for x in outliers_labels]
    else:
        table = None

    return fig, lgd, table


@SetProps
def sliding_window(data, window_size, ax_names=None, table_header=None,
                   title=None):
    """
    Creates a sliding window plot
    :param data:
    :param window_size:
    :param ax_names:
    :param table_header:
    """

    plt.style.use("ggplot")

    plt.rcParams["figure.figsize"] = (8, 6)

    data = np.array(data)
    x = np.arange(0, len(data) * window_size, window_size)

    table_data = []

    fig, ax = plt.subplots()

    for i in range(0, len(data)):
        ax.plot(x[i:i + 2], data[i:i + 2], color=cm.jet(data[i] / float(max(
            data)), 1), linewidth=2.)
        table_data.append([x[i], data[i]])

    ax.set_ylim([min(data) - min(data) * .1, max(data) + max(data) * .1])

    if ax_names:
        _add_labels(ax_names)

    if table_header:
        table = [table_header] + table_data

    else:
        table = table_data

    return fig, None, table


__author__ = "Diogo N. Silva"
__credits__ = ["Diogo N. Silva", "Tiago F. Jesus"]



try:
    from process import data
    from process.error_handling import KillByUser,\
        MultipleSequenceTypes, EmptyAlignment
    from process.sequence import AlignmentList, Alignment
    import orthomcl_pipeline as ortho_pipe
    from ortho import OrthomclToolbox as OrthoTool
except ImportError:
    from trifusion.process import data
    from trifusion.process.error_handling import KillByUser,\
        MultipleSequenceTypes, EmptyAlignment
    from trifusion.process.sequence import AlignmentList, Alignment
    import trifusion.orthomcl_pipeline as ortho_pipe
    from trifusion.ortho import OrthomclToolbox as OrthoTool

from os.path import join, basename
from collections import OrderedDict
from copy import deepcopy
import logging
import shutil
import cPickle as pickle
import time

import os
os.environ["KIVY_NO_ARGS"] = "1"

"""
Tasks that are executed in a background process are placed here and called
from the App class in TriFusion.py. The reason for this is that
multiprocessing in Windows does not rely on os.fork to spawn a new process
child with shared memory maps with the parent. Instead it relies on pickle to
transfer data. This reliance on pickle prevents application methods from being
sent to the background process because pickle cannot handle class methods.
The solution was to define the functions outside the App class and provide
them with the necessary arguments upon calling.
"""


def remove_tmp(temp_dir, sql_con):
    """
    Removes all temporary files in temp directory
    :param temp_dir: string, path to trifusion's temporary directory
    """

    # Give some time to child threads to exit
    time.sleep(1)

    # Close database connection
    sql_con.close()

    # Remove temporary files
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir)

    return 1


def load_proc(aln_list, file_list, nm, queue):
    try:
        if aln_list:
            aln_list.add_alignment_files(file_list,
                                         shared_namespace=nm)
            aln_obj = aln_list
        else:
            aln_obj = AlignmentList(file_list, shared_namespace=nm)

        queue.put(aln_obj)

    except MultipleSequenceTypes:
        nm.exception = "multiple_type"

    except IOError:
        return

    except KillByUser:
        return

    except Exception as e:
        logging.exception("Unexpected error when loading input data")
        print(e)


def get_stats_summary(dest, aln_list, active_file_set, active_taxa_set,
                        ns):
    """
    Runs the get_summary_stats method in the background and writes the output
    in a pickle file
    :param aln_list: AlignmentList object
    :param dest: temporary file where stats will be written
    :param active_file_set: list, with files to be included in summary
    statistics
    :param active_taxa_set: list, with taxa to be included in summary
    statistics
    """

    try:
        # Creating deepcopy to perform changes  without impacting main
        # attribute
        main_aln = deepcopy(aln_list)
        main_aln.set_database_connections(aln_list.cur, aln_list.con)
        # Update alignment object according to active file and taxa sets
        main_aln.update_active_alignments(active_file_set)
        main_aln.update_taxa_names(active_taxa_set)

        with open(join(dest, "stats.pc"), "wb") as fh_stats, \
                open(join(dest, "table.pc"), "wb") as fh_table:

            # Check if active data sets are not empty. If so, raise an
            # exception
            if main_aln.alignments == OrderedDict() or not main_aln.taxa_names:
                for fh in [fh_stats, fh_table]:
                    pickle.dump({"exception": "Alignment is empty after file "
                                              "and taxa filters"}, fh)
                return

            stats = main_aln.get_summary_stats(ns=ns)
            table = main_aln.get_gene_table_stats()
            pickle.dump(stats, fh_stats)
            pickle.dump(table, fh_table)

    except KillByUser:
        return


def background_process(f, ns, a):
    """
    Executes the func in the background and returns its value to the
    shared namespace
    :param f: bound method
    :param ns: shared namespace
    :param a: list arguments
    """
    try:
        if a:
            if "use_ns" in a:
                a.remove("use_ns")
                val = f(*a, ns=ns)
            else:
                val = f(*a)
        else:
            val = f()
        ns.val = val

    except KillByUser:
        return

    except IOError as e:
        print(e)

    except Exception:
        logging.exception("Unexpected exit in %s", f.__name__)
        ns.exception = True


def background_export_groups(f, nm, a):
    """
    Calls function f with a arguments
    :param f: bound method
    :param nm: shared namespace object to provide to bound method
    :param a: list, with arguments
    """
    try:
        f(*a, shared_namespace=nm)

    except KillByUser:
        pass

    except IOError as e:
        print(e)

    except:
        logging.exception("Unexpected error when exporting ortholog "
                          "groups")
        nm.exception = True


def orto_execution(nm, temp_dir, proteome_files, protein_min_len,
                   protein_max_stop, usearch_file, usearch_evalue,
                   usearch_threads, usearch_output, mcl_file, mcl_inflation,
                   ortholog_prefix, group_prefix, orto_max_gene,
                   orto_min_sp, sqldb, ortho_dir, usearch_db):
    """
    Executes all pipeline subprocesses sequentially and updates the
    Progess dialog label
    """

    try:
        nm.finished_tasks = []

        nm.task = "schema"
        ortho_pipe.install_schema(temp_dir)
        nm.finished_tasks = ["schema"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "adjust"
        ortho_pipe.adjust_fasta(proteome_files, ortho_dir, nm)
        nm.finished_tasks = ["schema", "adjust"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "filter"
        ortho_pipe.filter_fasta(protein_min_len, protein_max_stop,
                                usearch_db, ortho_dir, nm)
        nm.finished_tasks = ["schema", "adjust", "filter"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "usearch"
        ortho_pipe.allvsall_usearch(usearch_db, usearch_evalue, ortho_dir,
                                    usearch_threads, usearch_output,
                                    usearch_bin=usearch_file, nm=nm)
        nm.finished_tasks = ["schema", "adjust", "filter", "usearch"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "parse"
        ortho_pipe.blast_parser(usearch_output, ortho_dir,
                                db_dir=temp_dir, nm=nm)
        nm.finished_tasks = ["schema", "adjust", "filter", "usearch", "parse"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "pairs"
        ortho_pipe.pairs(temp_dir, nm=nm)
        ortho_pipe.dump_pairs(temp_dir, ortho_dir, nm=nm)
        nm.finished_tasks = ["schema", "adjust", "filter", "usearch", "parse",
                             "pairs"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "mcl"
        ortho_pipe.mcl(mcl_inflation, ortho_dir, mcl_file=mcl_file, nm=nm)
        nm.finished_tasks = ["schema", "adjust", "filter", "usearch", "parse",
                             "pairs", "mcl"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "dump"
        ortho_pipe.mcl_groups(mcl_inflation, ortholog_prefix, "1000",
                              group_prefix, ortho_dir, nm=nm)
        nm.finished_tasks = ["schema", "adjust", "filter", "usearch", "parse",
                             "pairs", "mcl", "dump"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "filter_groups"
        stats, groups_obj = ortho_pipe.export_filtered_groups(
            mcl_inflation,
            group_prefix,
            orto_max_gene,
            orto_min_sp,
            sqldb,
            join(ortho_dir, "backstage_files", usearch_db),
            temp_dir,
            ortho_dir, nm=nm)
        nm.finished_tasks = ["schema", "adjust", "filter", "usearch", "parse",
                             "pairs", "mcl", "dump", "filter_groups"]

        if nm.stop:
            raise KillByUser("")

        # stats is a dictionary containing the inflation value as
        #  key and a list with the orthologs as value
        nm.stats = stats
        nm.groups = groups_obj

    except KillByUser:
        return

    except IOError as e:
        nm.exception = str(e)
        print(e)
        return

    except Exception as e:
        logging.exception("Unexpected exit in Orthology search")
        nm.exception = str(e)


def update_active_fileset(aln_obj, set_name, file_list, file_groups,
                          filename_map):
    """
    This method is similar in purpose and functionality to the
    update_active_taxaset, but it updates the set of files. It should be
    used before the said method.
    :param aln_obj: The alignment object being used during execution of
    Process operations
    """
    # Determine the selected active taxa set from the dropdown menu
    if set_name == "All files":
        aln_obj.update_active_alignments([x for x in file_list])
        return aln_obj
    if set_name == "Active files":
        return aln_obj
    else:
        aln_obj.update_active_alignments(
            [filename_map[x] for x in file_groups[set_name]])
        return aln_obj


def update_active_taxaset(aln_obj, set_name, active_taxa_list, taxa_groups):
    """
    Do not use this method on the original self.alignment_list or
    self.active_alignment list, as it may cause unwanted permanent changes
    to the taxa set.
    This will take the complete taxa set from self.alignment_list.taxa_names
    and the currently active taxa set from self.active_taxa_list and remove
    the all taxa that are not present in the active taxa set from the
    alignment object passed as argument. If the lists are the same, no taxa
    will be removed
    """

    if set_name == "All taxa":
        return aln_obj
    if set_name == "Active taxa":
        tx_set = active_taxa_list
    else:
        tx_set = taxa_groups[set_name]

    # Update active taxa
    aln_obj.update_taxa_names(tx_set)
    return aln_obj


def process_execution(aln_list, file_set_name, file_list, file_groups,
                        filename_map,
                        taxa_set_name, active_taxa_list, ns, taxa_groups,
                        hap_prefix, secondary_operations, secondary_options,
                        missing_filter_settings, taxa_filter_settings,
                        codon_filter_settings, variation_filter_settings,
                        output_file, rev_infile, main_operations, zorro_suffix,
                        partitions_file, output_formats, create_partfile,
                        use_nexus_partitions, use_nexus_models,
                        phylip_truncate_name, output_dir, use_app_partitions,
                        consensus_type, ld_hat, ima2_params,
                        conversion_suffix):
    """
    Process execution function
    :param ns: Namespace object
    """

    def reverse_concatenation(aln):
        """
        Wrapper of the reverse concatenation operation
        :return: AlignmentList object
        """

        con = aln.con

        if not use_app_partitions:
            # Retrieve the alignment object that will be reverted. This
            # is done first in order to retrieve the length of the locus,
            # which is provided to the Partitions object for checking and
            # conversion of "." notation
            aln = aln.retrieve_alignment(rev_infile)

            # Instanciate Partitions object and set its length attribute
            partition_obj = data.Partitions()
            partition_obj.set_length(aln.locus_length)

            # In case the partitions file is badly formatted or invalid, the
            # exception will be returned by the read_from_file method.
            e = partition_obj.read_from_file(partitions_file)
            if e:
                ns.exception = {"InvalidPartitionFile": e.value}
                raise data.InvalidPartitionFile("")

            # If there are no issues with the partitions file, set the new
            # partitions
            aln.set_partitions(partition_obj)

        if aln.__class__.__name__ == "AlignmentList":
            aln = aln.reverse_concatenate(ns=ns)
        else:
            aln = aln.reverse_concatenate(db_con=con, ns=ns)

        return aln

    def filter_aln(aln, table_out="_filter"):
        """
        Wrapper for filtering operations, given an alignment object
        :param aln: AlignmentList object
        :param table_out: string. Suffix of the sqlite table for operations
        that change alignments
        """

        # Check if a minimum taxa representation was specified
        if secondary_options["gap_filter"]:
            if missing_filter_settings[1][0]:
                ns.main_msg = "Filter (minimum taxa)"
                aln.filter_min_taxa(missing_filter_settings[1][1], ns=ns)

        # Filter by taxa
        if secondary_options["taxa_filter"]:
            # Get taxa list from taxa groups
            ns.main_msg = "Filter (by taxa)"
            taxa_list = taxa_groups[taxa_filter_settings[1]]
            aln.filter_by_taxa(taxa_filter_settings[0], taxa_list, ns=ns)

        # Filter codon positions
        if secondary_options["codon_filter"]:
            ns.main_msg = "Filter (by codon)"
            aln.filter_codon_positions(codon_filter_settings,
                                       table_out=table_out, ns=ns)

        # Filter missing data
        if secondary_options["gap_filter"]:
            ns.main_msg = "Filter (by missing data)"
            if missing_filter_settings[0][0]:
                aln.filter_missing_data(missing_filter_settings[0][1],
                                        missing_filter_settings[0][2],
                                        table_out=table_out, ns=ns)

        # Filter variation
        if secondary_options["variation_filter"]:
            # Checks for variable site filter
            if variation_filter_settings[0] or variation_filter_settings[1]:
                ns.main_msg = "Filter (by variable sites)"
                aln.filter_segregating_sites(variation_filter_settings[0],
                                             variation_filter_settings[1],
                                             table_in=table_out, ns=ns)
            # Checks for informative site filter
            if variation_filter_settings[2] or variation_filter_settings[3]:
                ns.main_msg = "Filter (by informative sites)"
                aln.filter_informative_sites(variation_filter_settings[2],
                                             variation_filter_settings[3],
                                             table_in=table_out, ns=ns)

        # Pipe the information on the filtered alignments to the main process
        # only if it was applied a filter that changes the final alignments
        if set(aln.filtered_alignments.values()) != {None}:
            ns.filtered_alns = aln.filtered_alignments

        # Reset main label text
        ns.main_msg = None

        # Some filter configurations may result in empty final alignment
        # list. In such cases, return and issue warning
        if not aln.alignments:
            ns.exception = "EmptyAlignment"
            raise EmptyAlignment("Active alignment is empty")

        return aln

    def concatenation(aln, table_in=""):
        """
        Wrapper for concatenation operation
        :param aln: AlignmentList object
        """

        if secondary_options["zorro"]:
            ns.msg = "Concatenating ZORRO files"
            zorro_data = data.Zorro(aln, zorro_suffix)
            zorro_data.write_to_file(output_file)

        aln = aln.concatenate(alignment_name=basename(output_file),
                              table_in=table_in, ns=ns)

        # Sets the single alignment to True, for other method to be aware of
        # this
        ns.sa = True

        return aln

    def consensus(aln, table_out):
        """
        Wrapper of the consensus operation
        :param aln: AlignmentObject list
        """

        if secondary_options["consensus_single"]:
            if aln.__class__.__name__ == "AlignmentList":
                aln = aln.consensus(consensus_type=consensus_type,
                                    single_file=True,
                                    table_out=table_out,
                                    ns=ns)
                ns.sa = True
            else:
                aln.consensus(consensus_type=consensus_type,
                              table_out=table_out,
                              ns=ns)
        else:
            aln.consensus(consensus_type=consensus_type,
                          table_out=table_out,
                          ns=ns)

        return aln

    def writer(aln, filename=None, suffix_str="", conv_suffix="",
               table_suffix=None, table_name=None):
        """
        Wrapper for the output writing operations
        :param aln: AlignmentList object
        :param filename: string. If provided, it will overwrite the output_file
        :param suffix_str: string. Provides the suffix for the AlignmentList
        write_to_file method argument
        :param conv_suffix: string. Provides the suffix provided for
        the conversion of files. This suffix will allway precede the
        suffix_str, which is meant to apply suffixes specific to secondary
        operations.
        :param table_suffix: string. Suffix of the table from where the
        sequence data will be retrieved
        :param table_name: string. Name of the table from where the
        sequence data will be retrived
        """

        try:
            if filename:
                outfile = filename
            else:
                outfile = output_file

            # The output file(s) will only be written after all the required
            # operations have been concluded. The reason why there are two "if"
            # statement for "concatenation" is that the input alignments must
            # be concatenated before any other additional operations. If the
            # first if statement did not exist, then all additional options
            # would have to be manually written for both "conversion" and
            # "concatenation". As it is, when "concatenation", the aln_obj is
            # firstly converted into the concatenated alignment, and then all
            # additional operations are conducted in the same aln_obj

            if aln.__class__.__name__ == "Alignment":
                aln.write_to_file(output_formats,
                    outfile if outfile else join(output_dir, "consensus"),
                    interleave=secondary_options["interleave"],
                    partition_file=create_partfile,
                    use_charset=use_nexus_partitions,
                    phy_truncate_names=phylip_truncate_name,
                    ld_hat=ld_hat,
                    ima2_params=ima2_params,
                    use_nexus_models=use_nexus_models,
                    ns_pipe=ns,
                    table_suffix=table_suffix,
                    table_name=table_name)
            elif aln.__class__.__name__ == "AlignmentList":
                aln.write_to_file(
                    output_formats,
                    output_suffix=suffix_str,
                    conversion_suffix=conv_suffix,
                    interleave=secondary_options["interleave"],
                    partition_file=create_partfile,
                    output_dir=output_dir,
                    use_charset=use_nexus_partitions,
                    phy_truncate_names=phylip_truncate_name,
                    ld_hat=ld_hat,
                    ima2_params=ima2_params,
                    use_nexus_models=use_nexus_models,
                    ns_pipe=ns,
                    table_suffix=table_suffix,
                    table_name=table_name)

        except IOError as e:
            logging.exception(e)

    try:

        aln_object = deepcopy(aln_list)
        # Restore database connections, since they are broken during the
        # deepcopy operation
        aln_object.set_database_connections(aln_list.cur, aln_list.con)

        # Setting the alignment to use.
        # Update active file set of the alignment object
        aln_object = update_active_fileset(aln_object,
                                           file_set_name,
                                           file_list,
                                           file_groups,
                                           filename_map)

        # Update active taxa set of the alignment object
        main_aln = update_active_taxaset(aln_object, taxa_set_name,
                                           active_taxa_list,
                                           taxa_groups)

        ns.proc_files = len(aln_object.alignments)

        # Initialize attribute tha will store the number of filtered
        # alignments for reporting purposes
        ns.filtered_alns = None

        # The execution of the process module will begin with all the
        # operations on the main output alignment. Only after the main
        # output file has been created will the additional secondary output
        # files be processed.

        #####
        # Perform operations on MAIN OUTPUT
        #####

        # Set the suffix for the sqlite table harboring the main output
        # alignment if any of the secondary operations is specified
        if any(secondary_operations.values()):
            main_table = "main"
        else:
            main_table = ""

        # Reverse concatenation
        # Active table: Based on partition names
        if main_operations["reverse_concatenation"]:
            ns.task = "reverse_concatenation"
            main_aln = reverse_concatenation(main_aln)
            ns.finished_tasks.append("reverse_concatenation")

        # Filtering
        # Active table: * / *main
        if secondary_options["collapse_filter"] and not \
                secondary_options["collapse_file"]:
            ns.task = "collapse"
            # If the the collapse filter is active, perform this
            # filtering first. This is because the filter will allow 0% of
            # missing data, which will always be as stringent or more than any
            # missing data filter set.
            main_aln.filter_missing_data(0, 0, table_out=main_table, ns=ns)
            ns.finished_tasks.append("collapse_filter")

        # Active table: * / *main
        if secondary_operations["filter"] and not \
                secondary_options["filter_file"]:
            ns.task = "filter"
            main_aln = filter_aln(main_aln, table_out=main_table)
            ns.finished_tasks.append("filter")
        # Concatenation
        # Active table: concatenation
        if main_operations["concatenation"]:
            ns.task = "concatenation"
            main_aln = concatenation(main_aln, table_in=main_table)
            ns.finished_tasks.append("concatenation")
        # Collapsing
        # Active table: *main / concatenationmain
        if secondary_operations["collapse"] and not \
                secondary_options["collapse_file"]:
            ns.task = "collapse"
            main_aln.collapse(haplotype_name=hap_prefix, dest=output_dir,
                                table_out=main_table, ns=ns)
            ns.finished_tasks.append("collapse")
        # Gcoder
        # Active table: *main / concatenationmain
        if secondary_operations["gcoder"] and not \
                secondary_options["gcoder_file"]:
            ns.task = "gcoder"
            main_aln.code_gaps(table_out=main_table, ns=ns)
            ns.finished_tasks.append("gcoder")
        # Consensus
        # Active table: *main / concatenationmain / consensus
        if secondary_operations["consensus"] and not \
                secondary_options["consensus_file"]:
            ns.task = "consensus"
            main_aln = consensus(main_aln, table_out=main_table)
            ns.finished_tasks.append("consensus")

        # ### Guide on possible final tables
        # --- Base types
        # 1: Conversion -> * (main -- checks)
        # 2: Concatenation -> concatenation (main -- checks)
        # 3: Reverse concatenation -> * (main --checks)
        # --- Conversion and Reverse concatenation + secondary ops
        # 4: Collapse/Filter/Gcoder -> *main (sec --checks)
        # 5: Consensus -> *main (sec -- checks)
        # 6: Consensus (single file) -> consensus (main -- fallback)
        # --- Concatenation + secondary ops
        # 7: Filter (only) -> concatenation (main -- fallback)
        # 8: Collapse/gcoder -> concatenationmain (sec -- checks)
        # 9: Consensus -> concatenationmain (sec -- checks)
        # 10: Consensus (single file) -> consensus (main -- fallback)

        # NOTE ON TABLE NAMES
        # Some combinations of operations actually create a table_suffix
        # that does not exist in the database. This happens in cases 6, 7 and
        # 10. However, in these cases the main table(s) of the Alignment
        # object should be used, so we let the sequence fetching methods
        # fail to find the suggested table and fallback to the main table.

        ns.task = "write"
        writer(main_aln, conv_suffix=conversion_suffix,
               table_suffix=main_table)
        ns.finished_tasks.append("write")

        #####
        # Perform operations on ADDITIONAL OUTPUTS
        #####

        # Stores operations that must be performed before concatenation,
        # if it was specified
        before_conc = ["filter"]

        # Perform the filtering and consensus option separately, since these
        # must be done before concatenation
        for op in [x for x, y in secondary_operations.items() if
                   x in before_conc and y and
                   secondary_options["%s_file" % x]]:

            if op == "filter":

                ns.task = "filter"

                # Remove previous temporary tables
                aln_object.remove_tables(aln_object.get_tables())

                main_aln = deepcopy(aln_object)
                main_aln.set_database_connections(aln_list.cur, aln_list.con)

                ns.msg = "Creating additional filtered alignments(s)"
                suffix = "_filtered"
                main_aln = filter_aln(main_aln, table_out=suffix[1:])

            if main_operations["concatenation"] and \
                    main_aln.__class__.__name__ == "AlignmentList":

                filename = output_file + suffix

                ns.main_msg = "Concatenating"
                main_aln = concatenation(main_aln, table_in=suffix[1:])
                ns.main_msg = "Writing output"
                writer(main_aln, filename=filename, table_suffix=suffix[1:])
            else:
                ns.main_msg = "Writing output"
                writer(main_aln, suffix_str=suffix, table_suffix=suffix[1:],
                       conv_suffix=conversion_suffix)

        # Remove previous temporary tables
        aln_object.remove_tables(aln_object.get_tables())

        main_aln = deepcopy(aln_object)
        main_aln.set_database_connections(aln_list.cur, aln_list.con)

        concatenated = False
        ns.sa = False

        for op in [x for x, y in secondary_operations.items() if
                   x not in before_conc and y and
                   secondary_options["%s_file" % x]]:

            ns.task = op

            # Filter data for collapsing
            # if secondary_options["collapse_filter"] and op == "collapse":
                # ns.task = "collapse"
                # ns.main_msg = "Filtering for collapse"
                # main_aln.filter_missing_data(0, 0, ns=ns)

            if main_operations["concatenation"]:
                # If suffix was specified, it means that the filter was
                # ON. In that case, use that suffix in the table input
                # for concatenation.
                if op == "collapse" and secondary_options["collapse_filter"]:
                    pass
                else:
                    if not concatenated:
                        try:
                            main_aln = concatenation(aln_object)
                        except NameError:
                            main_aln = concatenation(aln_object)

                        concatenated = True

            if op == "consensus":

                suffix = "_consensus"
                main_aln = consensus(main_aln, table_out=suffix[1:])

            elif op == "gcoder":
                suffix = "_coded"
                main_aln.code_gaps(table_out=suffix[1:], ns=ns)

            elif op == "collapse":

                suffix = "_collapsed"

                # In this case, filter the unconcatenated alignment
                # and concatenate again
                if secondary_options["collapse_filter"]:
                    ns.main_msg = "Filtering for collapse"
                    aln_object.remove_tables(aln_object.get_tables())
                    aln_object.filter_missing_data(0, 0, ns=ns,
                                                   table_out=suffix[1:])
                    if main_operations["concatenation"]:
                        ns.main_msg = "Concatenation"
                        main_aln = aln_object.concatenate(table_in=suffix[1:],
                                                          ns=ns)

                ns.main_msg = "Collapse"
                main_aln.collapse(haplotype_name=hap_prefix,
                                  conversion_suffix=conversion_suffix,
                                  dest=output_dir,
                                  table_out=suffix[1:], ns=ns)

            ns.main_msg = "Writing output"
            if main_operations["concatenation"]:
                filename = output_file + suffix
                writer(main_aln, filename=filename, table_suffix=suffix[1:])
            else:
                writer(main_aln, suffix_str=suffix,
                       conv_suffix=conversion_suffix, table_suffix=suffix[1:])

        # Resets the taxa_names attribute of the aln_obj to include all taxa
        # aln_object.update_taxa_names(all_taxa=True)

    except KillByUser:
        pass

    except IOError as e:
        logging.exception(e)
        # Resets the taxa_names attribute of the aln_obj to include all taxa
        # aln_object.update_taxa_names(all_taxa=True)
        return

    except EmptyAlignment:
        logging.exception("Empty alignment")
        # Resets the taxa_names attribute of the aln_obj to include all taxa
        # aln_object.update_taxa_names(all_taxa=True)
        # ns.exception = "EmptyAlignment"

    except Exception as e:
        print(e)
        # Log traceback in case any unexpected error occurs. See
        # self.log_file for whereabouts of the traceback
        logging.exception("Unexpected exit in Process execution")
        # Resets the taxa_names attribute of the aln_obj to include all taxa
        # aln_object.update_taxa_names(all_taxa=True)
        if not hasattr(ns, "exception"):
            ns.exception = "Unknown"


def load_group_files(group_files, temp_dir, ns=None):
    og = OrthoTool.MultiGroupsLight(db_path=temp_dir,
                                    groups=group_files,
                                    ns=ns)
    return [og, og.filters]


def orto_update_filters(ortho_groups, gn_filter, sp_filter, excluded_taxa,
                        group_names=None, default=False):
    try:
        if group_names or group_names == []:
            ortho_groups.update_filters(gn_filter, sp_filter, excluded_taxa,
                                        group_names, default=default)
        else:
            ortho_groups.update_filters(gn_filter, sp_filter, excluded_taxa,
                                        default=default)
    except EOFError:
        pass
    return [ortho_groups]


def get_active_group(ortho_groups, old_active_group, active_group_name):

    if not old_active_group:
        active_group = ortho_groups.get_group(active_group_name)
    else:
        if active_group_name == old_active_group.name:
            return [None]
        else:
            active_group = ortho_groups.get_group(active_group_name)

    return [active_group]


def get_orto_data(active_group, plt_idx, filt, exclude_taxa):
    """
    Given a GroupLight object, this function will execute the method that
    corresponds to plt_idx to generate data.
    :param active_group: GroupLight object. Active group that will be used
    to plot data
    :param plt_idx: string. Identifier of the plot type that must have a
    correspondence in the method dictionary below
    :return plot_data: dictionary. Has the required data for the
    plotting methods to generate the plot.
    """

    # Store the plot generation method in a dictionary where keys are
    # the text attributes of the plot spinner and the values are
    # bound methods
    methods = {
        "Taxa distribution": active_group.bar_species_distribution,
        "Taxa coverage": active_group.bar_species_coverage,
        "Gene copy distribution": active_group.bar_genecopy_distribution,
        "Taxa gene copies": active_group.bar_genecopy_per_species
    }

    # Check for excluded taxa. If any have been provided and are different from
    # the ones already set in the GroupLight object, then update the taxa
    # filter.
    if (exclude_taxa and exclude_taxa != active_group.excluded_taxa) or \
            (exclude_taxa == [] and
             exclude_taxa != active_group.excluded_taxa):
        # The update_stats flag of the exclude_taxa method is set to True
        # to update the group summary stats. This is important for eventual
        # corrections to the ortholog cluster filters.
        active_group.exclude_taxa(exclude_taxa, True)

    if filt:
        # If filters AND excluded taxa have been provided, the first thing
        # to do is check whether the provided filters are still correct.
        # Excluding taxa may lead to changes in the maximum values of the
        # ortholog filters, and this needs to be corrected here
        if exclude_taxa and exclude_taxa != active_group.excluded_taxa:
            # Correct gene copy filter maximum
            gn_filt = filt[0] if filt[0] <= active_group.max_extra_copy \
                else active_group.max_extra_copy
            # Correct minimum taxa representation maximum
            sp_filt = filt[1] if \
                filt[1] <= len(active_group.species_list) \
                else len(active_group.species_list)
        else:
            # No taxa have been excluded this time, so we'll keep the provided
            # filters
            gn_filt, sp_filt = filt

        active_group.update_filters(gn_filt, sp_filt, True)

    plot_data = methods[plt_idx](filt=filt)

    return [plot_data]


def get_stats_data(aln_obj, stats_idx, active_file_set, active_taxa_set,
                     additional_args, ns=None):
    """
    Given an aln_obj, this function will execute the according method to
    generate plot data

    :param aln_obj: AlignmentObject
    :param stats_idx: string, identifier that maps to an AlignmentObject method
    :return: data for plot production
    """

    # Creating deepcopy to perform changes without impacting the main attribute
    main_aln = deepcopy(aln_obj)
    main_aln.set_database_connections(aln_obj.cur, aln_obj.con)

    # Update alignment object according to active file and taxa sets
    main_aln.update_active_alignments(active_file_set)
    main_aln.update_taxa_names(active_taxa_set)

    # Check if active data sets are not empty. If so, raise an exception
    if main_aln.alignments == OrderedDict() or not main_aln.taxa_names:
        return [EmptyAlignment("Active alignment is empty")]

    # List of gene specific idx. These plots only have one gene for the footer
    gene_specific = ["Pairwise sequence similarity gn"]

    if stats_idx in gene_specific:
        footer = [1, len(main_aln.taxa_names)]
    else:
        footer = [len(main_aln.alignments), len(main_aln.taxa_names)]

    methods = {"Gene occupancy": main_aln.gene_occupancy,
               "Distribution of missing data sp":
                   main_aln.missing_data_per_species,
               "Distribution of missing data":
                   main_aln.missing_data_distribution,
               "Distribution of missing orthologs":
                   main_aln.missing_genes_per_species,
               "Distribution of missing orthologs avg":
                   main_aln.missing_genes_average,
               "Distribution of sequence size":
                   main_aln.average_seqsize_per_species,
               "Distribution of sequence size all": main_aln.average_seqsize,
               "Cumulative distribution of missing genes":
                   main_aln.cumulative_missing_genes,
               "Proportion of nucleotides or residues":
                   main_aln.characters_proportion,
               "Proportion of nucleotides or residues sp":
                   main_aln.characters_proportion_per_species,
               "Pairwise sequence similarity": main_aln.sequence_similarity,
               "Pairwise sequence similarity sp":
                   main_aln.sequence_similarity_per_species,
               "Pairwise sequence similarity gn":
                   main_aln.sequence_similarity_gene,
               "Segregating sites": main_aln.sequence_segregation,
               "Segregating sites sp": main_aln.sequence_segregation_per_species,
               "Segregating sites gn": main_aln.sequence_segregation_gene,
               "Segregating sites prop": main_aln.sequence_segregation,
               "Alignment length/Polymorphism correlation":
                   main_aln.length_polymorphism_correlation,
               "Distribution of taxa frequency":
                   main_aln.taxa_distribution,
               "Allele Frequency Spectrum":
                   main_aln.allele_frequency_spectrum,
               "Allele Frequency Spectrum prop":
                   main_aln.allele_frequency_spectrum,
               "Allele Frequency Spectrum gn":
                   main_aln.allele_frequency_spectrum_gene,
               "Missing data outliers": main_aln.outlier_missing_data,
               "Missing data outliers sp": main_aln.outlier_missing_data_sp,
               "Segregating sites outliers": main_aln.outlier_segregating,
               "Segregating sites outliers sp": main_aln.outlier_segregating_sp,
               "Sequence size outliers sp": main_aln.outlier_sequence_size_sp,
               "Sequence size outliers": main_aln.outlier_sequence_size}

    if additional_args:
        plot_data = methods[stats_idx](ns=ns, **additional_args)
    else:
        plot_data = methods[stats_idx](ns)

    return [plot_data, footer]


__author__ = "Diogo N. Silva"



#!/usr/bin/env python2
#
#  Copyright 2012 Unknown <diogo@arch>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import time
import argparse
from argparse import RawDescriptionHelpFormatter
from os.path import join
import os

try:
    from process.base import print_col, RED, GREEN, YELLOW
    from TriSeq import CleanUp
    from ortho import OrthomclToolbox as OT
    from ortho import protein2dna
except ImportError:
    from trifusion.process.base import print_col, RED, GREEN, YELLOW
    from trifusion.TriSeq import CleanUp
    from trifusion.ortho import OrthomclToolbox as OT
    from trifusion.ortho import protein2dna


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Command line interface for "
                                     "the TriFusion Orthology explore module.")

    # Main execution
    main_exec = parser.add_argument_group("Main execution")
    main_exec.add_argument("-in", dest="infile", nargs="*",
                           help="Provide the OrthoMCL group file(s). Multiple "
                           "group files are allowed but with restricted "
                           "access to some options.")
    main_exec.add_argument("-o", dest="output_dir", required=True,
                           help="Output directory.")

    # Filter options
    filter_opts = parser.add_argument_group("Filter options")
    filter_opts.add_argument("-t1", "--gene-threshold", dest="gn_threshold",
                             type=int, help="Provide the "
                             "maximum threshold of gene copy numbers per "
                             "species that should be allowed. If the option "
                             "is not specified all gene copy values will be "
                             "allowed.")
    filter_opts.add_argument("-t2", "--species-threshold", dest="sp_threshold",
                             help="Provide the minimum "
                             "number of species that should be allowed. If "
                             "the option is not specified, there will be no "
                             "lower limit.")

    # Conversion options
    convert_opts = parser.add_argument_group("Conversion options")
    convert_opts.add_argument("-g2f", dest="groups2fasta",
                              action="store_const", const=True,
                              help="Retrieve the"
                              " protein sequences of each ortholog cluster to a"
                              "single file per cluster. This retrieval is "
                              "performed after filtering ortholog using "
                              "options '-t1' and '-t2'. The protein database "
                              "used in the ortholog search must be provided "
                              "using the --protein-db.")
    convert_opts.add_argument("--protein-db", dest="protein_db",
                              help="Path to the protein database in fasta "
                              "format. Can be either a single file containing"
                              " sequences for all taxa or individual files "
                              "for each taxon.")
    convert_opts.add_argument("-p2d", dest="protein2dna", help="Convert "
                              "unaligned protein sequence files into their "
                              "corresponding nucleotide sequences. A CDS "
                              "database must be provided using the --cds-db "
                              "option and the protein files must be provided "
                              "using the --protein-db option. A nucleotide "
                              "sequence file will be generated for each "
                              "protein sequence file provided in --protein-db.")
    convert_opts.add_argument("--cds-db", dest="dna_db", help="Path to the "
                              "CDS database in fasta format. Can be either a "
                              "single file containing sequences for all taxa "
                              "or individual files for each taxon.")

    # Plotting options
    plot_opts = parser.add_argument_group("Plotting options")
    plot_opts.add_argument("-s", dest="plots", nargs="*",
                           choices=["1", "2", "3", "4", "5"], help="Use the "
                           "available choices to perform statistical analyses "
                           "on group files. All analyses are performed on the"
                           " provided group files, before applying any filters."
                           "Note that most analyses require a"
                           " single group file as input. Choice description:"
                           "\n\t1: Compares the number of total and filtered "
                           "ortholog clusters between multiple group files.\n"
                           "\t2: [Single group required] Distribution of the "
                           "number of taxa across orthologs.\n\t3: [Single "
                           "group required] Plots the available and missing "
                           "data proportions per taxon.\n\t4: [Single group "
                           "required] Plots the cumulative number of multiple "
                           "gene copies per taxon.\n\t5: [Single group "
                           "required] Distribution of gene copy numbers across"
                           "orthologs.")

    output_opts = parser.add_argument_group("Output options")
    output_opts.add_argument("-e", dest="export", action="store_const",
                             const=True, help="Exports the filtered groups "
                             "into a new file.")

    arg = parser.parse_args()


def main_check():
    """
    Performs sanity checks to argument combinations
    """

    if arg.protein2dna and arg.infile:
        print_col("Group file operations are ignored when specifying "
                  "conversion options.", YELLOW, 3)

    # Check if protein and cds data bases are provided when required
    if arg.groups2fasta and not arg.protein_db:
        print_col("A protein database must be provided to convert group "
                  "files into sequence files using the --protein-db option. "
                  "Exiting.", RED, 3)

    if arg.protein2dna and (not arg.protein_db and not arg.dna_db):
        print_col("A CDS data base and protein sequence files must be "
                  "provided to convert protein sequences into nucleotide "
                  "sequences using the --cds-db and --protein-db options, "
                  "respectively. Exiting.", RED, 3)

    # Print warnings when trying to execute options that are not available to
    # multiple input group files
    if len(arg.infile) > 1:
        if arg.groups2fasta or arg.protein2dna:
            print_col("Conversion options are only available for single "
                      "group files input.", YELLOW, 3)

    if arg.groups2fasta and not arg.gn_threshold and not arg.sp_threshold:
        print_col("No filters have been specified for the conversion of "
                  "group files into protein sequences. This may result in a "
                  "very large number of output files.", YELLOW, 3)


@CleanUp
def main():

    print_col("Executing TriOrtho module at %s %s" % (
        time.strftime("%d/%m/%Y"), time.strftime("%I:%M:%S")), GREEN, 3)

    # Create tmp dir
    os.makedirs(".tmp")

    # Arguments
    groups_file = arg.infile
    output_dir = arg.output_dir

    # Create output directory
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    if arg.protein2dna:
        print_col("Converting protein sequences into nucleotide sequences",
                  GREEN, 3)
        # Create database
        print_col("Creating database", GREEN, 3)
        id_db = protein2dna.create_db(arg.dna_db, ".tmp")
        # Create query for USEARCH
        print_col("Creating query", GREEN, 3)
        query_db = protein2dna.create_query(arg.protein_db, ".tmp")
        # Execute search
        print_col("Executing search", GREEN, 3)
        protein2dna.pair_search(".tmp")
        pair_db = protein2dna.get_pairs(".tmp")
        # Convert files
        print_col("Converting files", GREEN, 3)
        protein2dna.convert_protein_file(pair_db, query_db, id_db, output_dir)
        return print_col("Protein to nucleotide conversion complete", GREEN, 3)

    gene_threshold = arg.gn_threshold
    species_threshold = arg.sp_threshold
    protein_db = arg.protein_db

    if len(groups_file) == 1:

        print_col("Parsing group file", GREEN, 3)
        group_file = groups_file[0]
        group_object = OT.GroupLight(group_file, gene_threshold,
                                     species_threshold)

        # Check for plotting options
        if arg.plots:

            plt_methods = {"2": [group_object.bar_species_distribution,
                                 "Species distribution"],
                           "3": [group_object.bar_species_coverage,
                                 "Species data coverage"],
                           "4": [group_object.bar_genecopy_per_species,
                                 "Gene copies per species"],
                           "5": [group_object.bar_genecopy_distribution,
                                 "Gene copy distribution"]}

            for i in arg.plots:
                if i == "1":
                    print_col("Plotting option 1 requires multiple group "
                              "files. Skipping.", YELLOW, 3)
                    continue

                # Generate plot data and file
                print_col("Generating plot for %s" % plt_methods[i][1],
                          GREEN, 3)
                plot_obj, _, table = plt_methods[i][0](dest=output_dir)

        # Export filtered group file
        if arg.export:
            print_col("Exporting filtered group file using %s maximum gene "
                      "copies and %s minimum taxa representation" %
                      (gene_threshold, species_threshold), GREEN, 3)
            group_object.export_filtered_group(dest=output_dir)
            print_col("Filtering complete.\nTotal orthologs: %s;\nAfter gene "
                      "filter: %s;\nAfter species filter: %s;\nAfter both "
                      "filters: %s" % (len(group_object.species_frequency),
                                       group_object.num_gene_compliant,
                                       group_object.num_species_compliant,
                                       group_object.all_compliant), GREEN, 3)

        if arg.groups2fasta:
            print_col("Exporting group file as protein sequence files",
                      GREEN, 3)
            # Set sqlite file
            sqldb = join(".tmp", "group2protein.db")
            group_object.retrieve_sequences(sqldb, protein_db, output_dir)

    else:
        print_col("Parsing %s group files" % len(groups_file), GREEN, 3)
        multiple_groups_object = OT.MultiGroupsLight(".tmp",
                                                     groups_file,
                                                     gene_threshold,
                                                     species_threshold)

        if arg.plots:

            for i in arg.plots:

                if i != "1":
                    print_col("Plotting option %s requires a single group "
                              "file as input. Skipping." % str(i), YELLOW, 3)
                    continue

                print_col("Generating plot for Multiple group comparison",
                          GREEN, 3)
                multiple_groups_object.update_filters(gene_threshold,
                                                      species_threshold)
                multiple_groups_object.bar_orthologs(dest=output_dir)

        if arg.export:

            for gname, gobj in multiple_groups_object:

                gname = os.path.basename(gname)
                output_file = os.path.splitext(gname)[0] + "_filtered.txt"

                print_col("Exporting group file %s using %s maximum gene "
                          "copies and %s minimum taxa representation" %
                          (gname, gene_threshold, species_threshold), GREEN, 3)
                gobj.export_filtered_group(output_file_name=output_file,
                                           dest=output_dir)
                print_col("Filtering complete for group file %s.\nTotal "
                          "orthologs: %s;\nAfter gene filter: %s;\nAfter "
                          "species filter: %s;\nAfter both filters: %s" %
                          (gname,
                           len(gobj.species_frequency),
                           gobj.num_gene_compliant,
                           gobj.num_species_compliant,
                           gobj.all_compliant), GREEN, 3)

main_check()
main()


__author__ = "Diogo N. Silva"
